# Unit test align_series_and_exog_multiseries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.exceptions import MissingValuesWarning
from skforecast.utils import check_preprocess_series
from skforecast.utils import check_preprocess_exog_multiseries
from skforecast.utils import align_series_and_exog_multiseries
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series_2
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series_as_dict
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog_as_dict
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog_as_dict_datetime
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog_predict


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_and_exog_None():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is None.
    """
    input_series_is_dict = isinstance(series, dict)
    series_dict, series_indexes = check_preprocess_series(series=series)

    exog_dict = {'1': None, '2': None}

    expected_series_dict = {
        '1': pd.Series(
                 data  = series['1'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series['2'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '2'
             ),
    }
    expected_exog_dict = {'1': None, '2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_exog_None():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and `exog` is None.
    """
    series_diff = series_2.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict = {'l1': None, 'l2': None}

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_2['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_2['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {'l1': None, 'l2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is None.
    """
    series_diff = series_2.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict = {'l1': None, 'l2': None}

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {'l1': None, 'l2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))


def test_output_align_series_and_exog_multiseries_when_series_is_dict_different_lengths_and_nans_in_between():
    """
    Test align_series_and_exog_multiseries when `series` is a dict 
    with series of different lengths and NaNs in between with `exog` is None.
    """
    series_diff = {
        'l1': series_as_dict['l1'].copy(),
        'l2': series_as_dict['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict = {'l1': None, 'l2': None}

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {'l1': None, 'l2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))


def test_output_align_series_and_exog_multiseries_when_input_series_is_DataFrame_and_exog_DataFrame():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is a pandas DataFrame.
    """
    input_series_is_dict = isinstance(series, dict)
    series_dict, series_indexes = check_preprocess_series(series=series)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['1', '2'],
                       exog                 = exog,
                       exog_dict            = {'1': None, '2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        '1': pd.Series(
                 data  = series['1'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series['2'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '2'
             ),
    }
    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog.to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
        '2': pd.DataFrame(
                 data    = exog.to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between_and_exog_DataFrame():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is a 
    pandas DataFrame with no datetime index.
    """
    series_diff = series.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['1', '2'],
                       exog                 = exog,
                       exog_dict            = {'1': None, '2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        '1': pd.Series(
                 data  = series_diff['1'].to_numpy()[-40:],
                 index = pd.RangeIndex(start=10, stop=50),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series_diff['2'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '2'
             ),
    }
    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog.to_numpy()[-40:],
                 index   = pd.RangeIndex(start=10, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
        '2': pd.DataFrame(
                 data    = exog.to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between_and_exog_DataFrame_datetime():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is a 
    pandas DataFrame with datetime index.
    """
    series_diff = series_2.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan

    exog_datetime = exog.copy()
    exog_datetime.index = pd.date_range(start='2000-01-01', periods=len(exog), freq='D')

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['l1', 'l2'],
                       exog                 = exog_datetime,
                       exog_dict            = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_datetime.to_numpy()[-40:],
                  index   = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_datetime.to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_input_series_is_DataFrame_and_exog_dict():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is a dict.
    """
    exog_dict = {
        '1': exog.copy(),
        '2': exog['exog_1'].copy()
    }

    input_series_is_dict = isinstance(series, dict)
    series_dict, series_indexes = check_preprocess_series(series=series)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['1', '2'],
                       exog                 = exog_dict,
                       exog_dict            = {'1': None, '2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        '1': pd.Series(
                 data  = series['1'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series['2'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '2'
             ),
    }
    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog.to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
        '2': pd.DataFrame(
                 data    = exog['exog_1'].to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1']
             ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between_and_exog_dict():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is a 
    dict with no datetime index.
    """
    series_diff = series.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan

    exog_dict = {
        '1': exog.copy(),
        '2': exog['exog_1'].copy()
    }

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['1', '2'],
                       exog                 = exog_dict,
                       exog_dict            = {'1': None, '2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        '1': pd.Series(
                 data  = series_diff['1'].to_numpy()[-40:],
                 index = pd.RangeIndex(start=10, stop=50),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series_diff['2'].to_numpy(),
                 index = pd.RangeIndex(start=0, stop=50),
                 name  = '2'
             ),
    }
    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog.to_numpy()[-40:],
                 index   = pd.RangeIndex(start=10, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
        '2': pd.DataFrame(
                 data    = exog['exog_1'].to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1']
             ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between_and_exog_dict_datetime():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is a 
    dict with datetime index.
    """
    series_diff = series_2.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['l1', 'l2'],
                       exog                 = exog_as_dict_datetime,
                       exog_dict            = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_as_dict_datetime['l1'].to_numpy()[-40:],
                  index   = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_as_dict_datetime['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_dict_and_exog_dict_datetime():
    """
    Test align_series_and_exog_multiseries when `series` is a dict and `exog` is 
    a dict. Datetime index is a must.
    """
    input_series_is_dict = isinstance(series_as_dict, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_as_dict)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['l1', 'l2'],
                       exog                 = exog_as_dict_datetime,
                       exog_dict            = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_as_dict['l1'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_as_dict['l1']), freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_as_dict['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_as_dict['l2']), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_as_dict_datetime['l1'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog_as_dict_datetime['l1']), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_as_dict_datetime['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog_as_dict_datetime['l2']), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_dict_different_lengths_and_nans_in_between_and_exog_dict_datetime():
    """
    Test align_series_and_exog_multiseries when `series` is a dict 
    with series of different lengths and NaNs in between with `exog` is a 
    dict with datetime index.
    """
    series_diff = {
        'l1': series_as_dict['l1'].copy(),
        'l2': series_as_dict['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['l1', 'l2'],
                       exog                 = exog_as_dict_datetime,
                       exog_dict            = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict          = series_dict,
                                 input_series_is_dict = input_series_is_dict,
                                 exog_dict            = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_as_dict_datetime['l1'].to_numpy()[-40:],
                  index   = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_as_dict_datetime['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_dict_and_length_intersection_with_exog_is_0():
    """
    Test align_series_and_exog_multiseries when `series` is a dict and the
    intersection with `exog` is 0.
    """
    series_diff = {
        'l1': series_as_dict['l1'].copy(),
        'l2': series_as_dict['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_no_intersection = {
        'l1': exog_as_dict_datetime['l1'].copy(),
        'l2': exog_as_dict_datetime['l2'].copy()
    }
    exog_no_intersection['l1'].index = pd.date_range(
        start='2001-01-01', periods=len(exog_no_intersection['l1']), freq='D'
    )

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['l1', 'l2'],
                       exog                 = exog_no_intersection,
                       exog_dict            = {'l1': None, 'l2': None}
                   )

    warn_msg = re.escape(
        ("Series 'l1' and its `exog` do not have the same index. "
         "All exog values will be NaN for the period of the series.")
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        series_dict, exog_dict = align_series_and_exog_multiseries(
                                     series_dict          = series_dict,
                                     input_series_is_dict = input_series_is_dict,
                                     exog_dict            = exog_dict
                                 )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = np.nan,
                  index   = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_2': object}),
        'l2': pd.DataFrame(
                  data    = exog_as_dict_datetime['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  columns = ['exog_1']
              ),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_dict_and_different_length_intersection_with_exog():
    """
    Test align_series_and_exog_multiseries when `series` is a dict and the
    intersection with `exog` do not have the same length as series.
    """
    series_diff = {
        'l1': series_as_dict['l1'].copy(),
        'l2': series_as_dict['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    input_series_is_dict = isinstance(series_diff, dict)
    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_half_intersection = {
        'l1': exog_as_dict_datetime['l1'].copy(),
        'l2': exog_as_dict_datetime['l2'].copy()
    }
    exog_half_intersection['l1'].iloc[20:23, :] = np.nan
    exog_half_intersection['l1'].index = pd.date_range(
        start='2000-01-15', periods=len(exog_half_intersection['l1']), freq='D'
    )
    exog_half_intersection['l2'].iloc[:10] = np.nan
    exog_half_intersection['l2'].iloc[33:49] = np.nan

    exog_dict, _ = check_preprocess_exog_multiseries(
                       input_series_is_dict = input_series_is_dict,
                       series_indexes       = series_indexes,
                       series_col_names     = ['l1', 'l2'],
                       exog                 = exog_half_intersection,
                       exog_dict            = {'l1': None, 'l2': None}
                   )

    warn_msg = re.escape(
        ("Series 'l1' and its `exog` do not have the same length. "
         "Exog values will be NaN for the not matched period of the series.")
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        series_dict, exog_dict = align_series_and_exog_multiseries(
                                     series_dict          = series_dict,
                                     input_series_is_dict = input_series_is_dict,
                                     exog_dict            = exog_dict
                                 )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_2)-n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_2), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = np.vstack((
                                np.full((4, len(['exog_1', 'exog_2'])), np.nan), 
                                exog_half_intersection['l1'].to_numpy()[:40-4]
                            )),
                  index   = pd.date_range(start='2000-01-11', periods=len(expected_series_dict['l1']), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float, 'exog_2': object}),
        'l2': pd.DataFrame(
                  data    = exog_half_intersection['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog_half_intersection['l2']), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


################################################################################
#                            ForecasterAutoreg                                 #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from typing import Union, Tuple, Optional, Callable
import warnings
import logging
import sys
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import KBinsDiscretizer
from sklearn.base import clone
import inspect

import skforecast
from ..ForecasterBase import ForecasterBase
from ..utils import initialize_lags
from ..utils import initialize_weights
from ..utils import check_select_fit_kwargs
from ..utils import check_y
from ..utils import check_exog
from ..utils import get_exog_dtypes
from ..utils import check_exog_dtypes
from ..utils import check_predict_input
from ..utils import check_interval
from ..utils import preprocess_y
from ..utils import preprocess_last_window
from ..utils import preprocess_exog
from ..utils import expand_index
from ..utils import transform_series
from ..utils import transform_dataframe
from ..preprocessing import TimeSeriesDifferentiator

logging.basicConfig(
    format = '%(name)-10s %(levelname)-5s %(message)s', 
    level  = logging.INFO,
)


class ForecasterAutoreg(ForecasterBase):
    """
    This class turns any regressor compatible with the scikit-learn API into a
    recursive autoregressive (multi-step) forecaster.
    
    Parameters
    ----------
    regressor : regressor or pipeline compatible with the scikit-learn API
        An instance of a regressor or pipeline compatible with the scikit-learn API
    lags : int, list, numpy ndarray, range
        Lags used as predictors. Index starts at 1, so lag 1 is equal to t-1. 
    
        - `int`: include lags from 1 to `lags` (included).
        - `list`, `1d numpy ndarray` or `range`: include only lags present in 
        `lags`, all elements must be int.
    transformer_y : object transformer (preprocessor), default `None`
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API with methods: fit, transform, fit_transform and inverse_transform.
        ColumnTransformers are not allowed since they do not have inverse_transform method.
        The transformation is applied to `y` before training the forecaster. 
    transformer_exog : object transformer (preprocessor), default `None`
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API. The transformation is applied to `exog` before training the
        forecaster. `inverse_transform` is not available when using ColumnTransformers.
    weight_func : Callable, default `None`
        Function that defines the individual weights for each sample based on the
        index. For example, a function that assigns a lower weight to certain dates.
        Ignored if `regressor` does not have the argument `sample_weight` in its `fit`
        method. The resulting `sample_weight` cannot have negative values.
    differentiation : int, default `None`
        Order of differencing applied to the time series before training the forecaster.
        If `None`, no differencing is applied. The order of differentiation is the number
        of times the differencing operation is applied to a time series. Differencing
        involves computing the differences between consecutive data points in the series.
        Differentiation is reversed in the output of `predict()` and `predict_interval()`.
        **WARNING: This argument is newly introduced and requires special attention. It
        is still experimental and may undergo changes.**
        **New in version 0.10.0**
    fit_kwargs : dict, default `None`
        Additional arguments to be passed to the `fit` method of the regressor.
    binner_kwargs : dict, default `None`
        Additional arguments to pass to the `KBinsDiscretizer` used to discretize 
        the residuals into k bins according to the predicted values associated 
        with each residual. The `encode' argument is always set to 'ordinal' 
        and `dtype' to np.float64.
        **New in version 0.12.0**
    forecaster_id : str, int, default `None`
        Name used as an identifier of the forecaster.
    
    Attributes
    ----------
    regressor : regressor or pipeline compatible with the scikit-learn API
        An instance of a regressor or pipeline compatible with the scikit-learn API.
    lags : numpy ndarray
        Lags used as predictors.
    transformer_y : object transformer (preprocessor)
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API with methods: fit, transform, fit_transform and inverse_transform.
        ColumnTransformers are not allowed since they do not have inverse_transform method.
        The transformation is applied to `y` before training the forecaster.
    transformer_exog : object transformer (preprocessor)
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API. The transformation is applied to `exog` before training the
        forecaster. `inverse_transform` is not available when using ColumnTransformers.
    weight_func : Callable
        Function that defines the individual weights for each sample based on the
        index. For example, a function that assigns a lower weight to certain dates.
        Ignored if `regressor` does not have the argument `sample_weight` in its `fit`
        method. The resulting `sample_weight` cannot have negative values.
    differentiation : int, default `None`
        Order of differencing applied to the time series before training the forecaster.
        If `None`, no differencing is applied. The order of differentiation is the number
        of times the differencing operation is applied to a time series. Differencing
        involves computing the differences between consecutive data points in the series.
        Differentiation is reversed in the output of `predict()` and `predict_interval()`.
        **WARNING: This argument is newly introduced and requires special attention. It
        is still experimental and may undergo changes.**
        **New in version 0.10.0**
    binner : sklearn.preprocessing.KBinsDiscretizer
        `KBinsDiscretizer` used to discretize residuals into k bins according 
        to the predicted values associated with each residual.
        **New in version 0.12.0**
    binner_kwargs : dict
        Additional arguments to pass to the `KBinsDiscretizer` used to discretize 
        the residuals into k bins according to the predicted values associated 
        with each residual. The `encode' argument is always set to 'ordinal' 
        and `dtype' to np.float64.
        **New in version 0.12.0**
    source_code_weight_func : str
        Source code of the custom function used to create weights.
    differentiation : int
        Order of differencing applied to the time series before training the 
        forecaster.
    differentiator : TimeSeriesDifferentiator
        Skforecast object used to differentiate the time series.
    max_lag : int
        Maximum value of lag included in `lags`.
    window_size : int
        Size of the window needed to create the predictors. It is equal to `max_lag`.
    window_size_diff : int
        Size of the window extended by the order of differentiation. When using
        differentiation, the `window_size` is increased by the order of differentiation
        so that the predictors can be created correctly.
    last_window : pandas Series
        This window represents the most recent data observed by the predictor
        during its training phase. It contains the values needed to predict the
        next step immediately after the training data. These values are stored
        in the original scale of the time series before undergoing any transformations
        or differentiation. When `differentiation` parameter is specified, the
        dimensions of the `last_window` are expanded as many values as the order
        of differentiation. For example, if `lags` = 7 and `differentiation` = 1,
        `last_window` will have 8 values.
    index_type : type
        Type of index of the input used in training.
    index_freq : str
        Frequency of Index of the input used in training.
    training_range : pandas Index
        First and last values of index of the data used during training.
    included_exog : bool
        If the forecaster has been trained using exogenous variable/s.
    exog_type : type
        Type of exogenous data (pandas Series or DataFrame) used in training.
    exog_dtypes : dict
        Type of each exogenous variable/s used in training. If `transformer_exog` 
        is used, the dtypes are calculated after the transformation.
    exog_col_names : list
        Names of the exogenous variables used during training.
    X_train_col_names : list
        Names of columns of the matrix created internally for training.
    fit_kwargs : dict
        Additional arguments to be passed to the `fit` method of the regressor.
    in_sample_residuals : numpy ndarray
        Residuals of the model when predicting training data. Only stored up to
        1000 values. If `transformer_y` is not `None`, residuals are stored in the
        transformed scale.
    in_sample_residuals_by_bin : dict
        In sample residuals binned according to the predicted value each residual
        is associated with. Only stored up to 200 values per bin. If `transformer_y`
        is not `None`, residuals are stored in the transformed scale.
        **New in version 0.12.0**
    out_sample_residuals : numpy ndarray
        Residuals of the model when predicting non training data. Only stored
        up to 1000 values. If `transformer_y` is not `None`, residuals
        are assumed to be in the transformed scale. Use `set_out_sample_residuals` 
        method to set values.
    out_sample_residuals_by_bin : dict
        Out of sample residuals binned according to the predicted value each residual
        is associated with. Only stored up to 200 values per bin. If `transformer_y`
        is not `None`, residuals are assumed to be in the transformed scale.
        **New in version 0.12.0**
    fitted : bool
        Tag to identify if the regressor has been fitted (trained).
    creation_date : str
        Date of creation.
    fit_date : str
        Date of last fit.
    skforecast_version : str
        Version of skforecast library used to create the forecaster.
    python_version : str
        Version of python used to create the forecaster.
    forecaster_id : str, int
        Name used as an identifier of the forecaster.
    
    """
    
    def __init__(
        self,
        regressor: object,
        lags: Union[int, np.ndarray, list],
        transformer_y: Optional[object]=None,
        transformer_exog: Optional[object]=None,
        weight_func: Optional[Callable]=None,
        differentiation: Optional[int]=None,
        fit_kwargs: Optional[dict]=None,
        binner_kwargs: Optional[dict]=None,
        forecaster_id: Optional[Union[str, int]]=None
    ) -> None:
        
        self.regressor                   = regressor
        self.transformer_y               = transformer_y
        self.transformer_exog            = transformer_exog
        self.weight_func                 = weight_func
        self.source_code_weight_func     = None
        self.differentiation             = differentiation
        self.differentiator              = None
        self.last_window                 = None
        self.index_type                  = None
        self.index_freq                  = None
        self.training_range              = None
        self.included_exog               = False
        self.exog_type                   = None
        self.exog_dtypes                 = None
        self.exog_col_names              = None
        self.X_train_col_names           = None
        self.in_sample_residuals         = None
        self.out_sample_residuals        = None
        self.in_sample_residuals_by_bin  = None
        self.out_sample_residuals_by_bin = None
        self.fitted                      = False
        self.creation_date               = pd.Timestamp.today().strftime('%Y-%m-%d %H:%M:%S')
        self.fit_date                    = None
        self.skforecast_version          = skforecast.__version__
        self.python_version              = sys.version.split(" ")[0]
        self.forecaster_id               = forecaster_id
       
        self.lags = initialize_lags(type(self).__name__, lags)
        self.max_lag = max(self.lags)
        self.window_size = self.max_lag
        self.window_size_diff = self.max_lag
        
        self.binner_kwargs = binner_kwargs
        if binner_kwargs is None:
            self.binner_kwargs = {
                'n_bins': 15, 'encode': 'ordinal', 'strategy': 'quantile',
                'subsample': 10000, 'random_state': 789654, 'dtype': np.float64
            }
        else:
            self.binner_kwargs = binner_kwargs
            self.binner_kwargs['encode'] = 'ordinal'
            self.binner_kwargs['dtype'] = np.float64
        self.binner = KBinsDiscretizer(**self.binner_kwargs)
        self.binner_intervals = None

        if self.differentiation is not None:
            if not isinstance(differentiation, int) or differentiation < 1:
                raise ValueError(
                    (f"Argument `differentiation` must be an integer equal to or "
                     f"greater than 1. Got {differentiation}.")
                )
            self.window_size_diff += self.differentiation
            self.differentiator = TimeSeriesDifferentiator(order=self.differentiation)
            
        self.weight_func, self.source_code_weight_func, _ = initialize_weights(
            forecaster_name = type(self).__name__, 
            regressor       = regressor, 
            weight_func     = weight_func, 
            series_weights  = None
        )

        self.fit_kwargs = check_select_fit_kwargs(
                              regressor  = regressor,
                              fit_kwargs = fit_kwargs
                          )


    def __repr__(
        self
    ) -> str:
        """
        Information displayed when a ForecasterAutoreg object is printed.
        """

        if isinstance(self.regressor, Pipeline):
            name_pipe_steps = tuple(name + "__" for name in self.regressor.named_steps.keys())
            params = {key : value for key, value in self.regressor.get_params().items() \
                      if key.startswith(name_pipe_steps)}
        else:
            params = self.regressor.get_params(deep=True)

        info = (
            f"{'=' * len(type(self).__name__)} \n"
            f"{type(self).__name__} \n"
            f"{'=' * len(type(self).__name__)} \n"
            f"Regressor: {self.regressor} \n"
            f"Lags: {self.lags} \n"
            f"Transformer for y: {self.transformer_y} \n"
            f"Transformer for exog: {self.transformer_exog} \n"
            f"Window size: {self.window_size} \n"
            f"Weight function included: {True if self.weight_func is not None else False} \n"
            f"Differentiation order: {self.differentiation} \n"
            f"Exogenous included: {self.included_exog} \n"
            f"Type of exogenous variable: {self.exog_type} \n"
            f"Exogenous variables names: {self.exog_col_names} \n"
            f"Training range: {self.training_range.to_list() if self.fitted else None} \n"
            f"Training index type: {str(self.index_type).split('.')[-1][:-2] if self.fitted else None} \n"
            f"Training index frequency: {self.index_freq if self.fitted else None} \n"
            f"Regressor parameters: {params} \n"
            f"fit_kwargs: {self.fit_kwargs} \n"
            f"Creation date: {self.creation_date} \n"
            f"Last fit date: {self.fit_date} \n"
            f"Skforecast version: {self.skforecast_version} \n"
            f"Python version: {self.python_version} \n"
            f"Forecaster id: {self.forecaster_id} \n"
        )

        return info

    
    def _create_lags(
        self, 
        y: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Transforms a 1d array into a 2d array (X) and a 1d array (y). Each row
        in X is associated with a value of y and it represents the lags that
        precede it.
        
        Notice that, the returned matrix X_data, contains the lag 1 in the first
        column, the lag 2 in the second column and so on.
        
        Parameters
        ----------
        y : numpy ndarray
            1d numpy ndarray Training time series.

        Returns
        -------
        X_data : numpy ndarray
            2d numpy ndarray with the lagged values (predictors). 
            Shape: (samples - max(self.lags), len(self.lags))
        y_data : numpy ndarray
            1d numpy ndarray with the values of the time series related to each 
            row of `X_data`. 
            Shape: (samples - max(self.lags), )
        
        """
          
        n_splits = len(y) - self.max_lag
        if n_splits <= 0:
            raise ValueError(
                (f"The maximum lag ({self.max_lag}) must be less than the length "
                 f"of the series ({len(y)}).")
            )
        
        X_data = np.full(shape=(n_splits, len(self.lags)), fill_value=np.nan, dtype=float)

        for i, lag in enumerate(self.lags):
            X_data[:, i] = y[self.max_lag - lag: -lag]

        y_data = y[self.max_lag:]
        
        return X_data, y_data


    def create_train_X_y(
        self,
        y: pd.Series,
        exog: Optional[Union[pd.Series, pd.DataFrame]]=None
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Create training matrices from univariate time series and exogenous
        variables.
        
        Parameters
        ----------
        y : pandas Series
            Training time series.
        exog : pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `y` and their indexes must be aligned.

        Returns
        -------
        X_train : pandas DataFrame
            Training values (predictors).
            Shape: (len(y) - self.max_lag, len(self.lags))
        y_train : pandas Series
            Values (target) of the time series related to each row of `X_train`.
            Shape: (len(y) - self.max_lag, )
        
        """
        
        check_y(y=y)
        fit_transformer = False if self.fitted else True
        y = transform_series(
                series            = y,
                transformer       = self.transformer_y,
                fit               = fit_transformer,
                inverse_transform = False
            )
        y_values, y_index = preprocess_y(y=y)

        if self.differentiation is not None:
            if not self.fitted:
                y_values = self.differentiator.fit_transform(y_values)
            else:
                differentiator = clone(self.differentiator)
                y_values = differentiator.fit_transform(y_values)
        
        if exog is not None:
            
            check_exog(exog=exog, allow_nan=True)
            if len(exog) != len(y):
                raise ValueError(
                    (f"`exog` must have same number of samples as `y`. "
                     f"length `exog`: ({len(exog)}), length `y`: ({len(y)})")
                )
            
            if isinstance(exog, pd.Series):
                exog = transform_series(
                           series            = exog,
                           transformer       = self.transformer_exog,
                           fit               = fit_transformer,
                           inverse_transform = False
                       )
            else:
                exog = transform_dataframe(
                           df                = exog,
                           transformer       = self.transformer_exog,
                           fit               = fit_transformer,
                           inverse_transform = False
                       )
            
            check_exog_dtypes(exog, call_check_exog=True)

            _, exog_index = preprocess_exog(exog=exog, return_values=False)
            if not (exog_index[:len(y_index)] == y_index).all():
                raise ValueError(
                    ("Different index for `y` and `exog`. They must be equal "
                     "to ensure the correct alignment of values.")
                )
        
        X_train, y_train = self._create_lags(y=y_values)
        X_train_col_names = [f"lag_{i}" for i in self.lags]
        X_train = pd.DataFrame(
                      data    = X_train,
                      columns = X_train_col_names,
                      index   = y_index[self.max_lag: ]
                  )
        
        if exog is not None:
            # The first `self.max_lag` positions have to be removed from exog
            # since they are not in X_train.
            exog_to_train = exog.iloc[self.max_lag:, ]
            exog_to_train.index = exog_index[self.max_lag:]
            X_train = pd.concat((X_train, exog_to_train), axis=1)
        
        # TODO: move self to fit method and make X_train_col_names a return
        if not self.fitted:
            self.X_train_col_names = X_train.columns.to_list()
        
        y_train = pd.Series(
                      data  = y_train,
                      index = y_index[self.max_lag: ],
                      name  = 'y'
                  )

        if self.differentiation is not None:
            X_train = X_train.iloc[self.differentiation: ]
            y_train = y_train.iloc[self.differentiation: ]
                        
        return X_train, y_train


    def create_sample_weights(
        self,
        X_train: pd.DataFrame,
    )-> np.ndarray:
        """
        Crate weights for each observation according to the forecaster's attribute
        `weight_func`.

        Parameters
        ----------
        X_train : pandas DataFrame
            Dataframe created with the `create_train_X_y` method, first return.

        Returns
        -------
        sample_weight : numpy ndarray
            Weights to use in `fit` method.

        """

        sample_weight = None

        if self.weight_func is not None:
            sample_weight = self.weight_func(X_train.index)

        if sample_weight is not None:
            if np.isnan(sample_weight).any():
                raise ValueError(
                    "The resulting `sample_weight` cannot have NaN values."
                )
            if np.any(sample_weight < 0):
                raise ValueError(
                    "The resulting `sample_weight` cannot have negative values."
                )
            if np.sum(sample_weight) == 0:
                raise ValueError(
                    ("The resulting `sample_weight` cannot be normalized because "
                     "the sum of the weights is zero.")
                )

        return sample_weight

        
    def fit(
        self,
        y: pd.Series,
        exog: Optional[Union[pd.Series, pd.DataFrame]]=None,
        store_last_window: bool=True,
        store_in_sample_residuals: bool=True
    ) -> None:
        """
        Training Forecaster.

        Additional arguments to be passed to the `fit` method of the regressor 
        can be added with the `fit_kwargs` argument when initializing the forecaster.
        
        Parameters
        ----------
        y : pandas Series
            Training time series.
        exog : pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `y` and their indexes must be aligned so
            that y[i] is regressed on exog[i].
        store_last_window : bool, default `True`
            Whether or not to store the last window of training data.
        store_in_sample_residuals : bool, default `True`
            If `True`, in-sample residuals will be stored in the forecaster object
            after fitting.

        Returns
        -------
        None
        
        """
        
        # Reset values in case the forecaster has already been fitted.
        self.index_type          = None
        self.index_freq          = None
        self.last_window         = None
        self.included_exog       = False
        self.exog_type           = None
        self.exog_dtypes         = None
        self.exog_col_names      = None
        self.X_train_col_names   = None
        self.in_sample_residuals = None
        self.fitted              = False
        self.training_range      = None

        X_train, y_train = self.create_train_X_y(y=y, exog=exog)
        sample_weight = self.create_sample_weights(X_train=X_train)

        if sample_weight is not None:
            self.regressor.fit(X=X_train, y=y_train, sample_weight=sample_weight,
                               **self.fit_kwargs)
        else:
            self.regressor.fit(X=X_train, y=y_train, **self.fit_kwargs)

        self.fitted = True
        self.fit_date = pd.Timestamp.today().strftime('%Y-%m-%d %H:%M:%S')
        self.training_range = preprocess_y(y=y, return_values=False)[1][[0, -1]]
        self.index_type = type(X_train.index)
        if isinstance(X_train.index, pd.DatetimeIndex):
            self.index_freq = X_train.index.freqstr
        else: 
            self.index_freq = X_train.index.step

        if exog is not None:
            self.included_exog = True
            self.exog_type = type(exog)
            self.exog_dtypes = get_exog_dtypes(exog=exog)
            self.exog_col_names = \
                 exog.columns.to_list() if isinstance(exog, pd.DataFrame) else exog.name
        
        # This is done to save time during fit in functions such as backtesting()
        if store_in_sample_residuals:
            in_sample_predictions = pd.Series(
                                        data  = self.regressor.predict(X_train),
                                        index = X_train.index
                                    )
            self._binning_in_sample_residuals(
                y_true = y_train,
                y_pred = in_sample_predictions
            )
        
        # The last time window of training data is stored so that lags needed as
        # predictors in the first iteration of `predict()` can be calculated. It
        # also includes the values need to calculate the diferenctiation.
        if store_last_window:
            self.last_window = y.iloc[-self.window_size_diff:].copy()


    def _binning_in_sample_residuals(
        self,
        y_true: pd.Series,
        y_pred: pd.Series,
    ) -> None:
        """
        Binning residuals according to the predicted value each residual is
        associated with. First a sklearn.preprocessing.KBinsDiscretizer
        is fitted to the predicted values. Then, residuals are binned according
        to the predicted value each residual is associated with. Residuals are
        stored in the forecaster object as `in_sample_residuals` and
        `in_sample_residuals_by_bin`. Only up to 200 residuals are stored per bin.

        Parameters
        ----------
        y_true : pandas Series
            True values of the time series.
        y_pred : pandas Series
            Predicted values of the time series.     

        Returns
        -------
        None
        
        """

        y_pred = y_pred.rename('prediction')
        residuals = (y_true - y_pred).rename('residual')
        data = pd.merge(
                   residuals,
                   y_pred,
                   left_index  = True,
                   right_index = True
               )
        self.binner.fit(data[['prediction']].to_numpy())
        data['bin'] = self.binner.transform(data[['prediction']].to_numpy()).astype(int)
        self.in_sample_residuals_by_bin = (
            data.groupby('bin')['residual'].apply(np.array).to_dict()
        )

        # Only up to 200 residuals are stored per bin
        for k, v in self.in_sample_residuals_by_bin.items():
            # TODO: Include `random_state` in fit method to allow the user 
            # change the residual sample stored.
            rng = np.random.default_rng(seed=95123)
            if len(v) > 200:
                sample = rng.choice(a=v, size=200, replace=False)
                self.in_sample_residuals_by_bin[k] = sample

        self.in_sample_residuals = np.concatenate(list(
            self.in_sample_residuals_by_bin.values()
        ))

        self.binner_intervals = {
            i: (
                self.binner.bin_edges_[0][i],
                (
                    self.binner.bin_edges_[0][i + 1]
                    if i + 1 < len(self.binner.bin_edges_[0])
                    else None
                ),
            )
            for i in range(len(self.binner.bin_edges_[0]) - 1)
        }


    def _recursive_predict(
        self,
        steps: int,
        last_window: np.ndarray,
        exog: Optional[np.ndarray]=None
    ) -> np.ndarray:
        """
        Predict n steps ahead. It is an iterative process in which, each prediction,
        is used as a predictor for the next step.
        
        Parameters
        ----------
        steps : int
            Number of future steps predicted.
        last_window : numpy ndarray
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
        exog : numpy ndarray, default `None`
            Exogenous variable/s included as predictor/s.

        Returns
        -------
        predictions : numpy ndarray
            Predicted values.
        
        """

        predictions = np.full(shape=steps, fill_value=np.nan)

        for i in range(steps):
            X = last_window[-self.lags].reshape(1, -1)
            if exog is not None:
                X = np.column_stack((X, exog[i, ].reshape(1, -1)))
            with warnings.catch_warnings():
                # Suppress scikit-learn warning: "X does not have valid feature names,
                # but NoOpTransformer was fitted with feature names".
                warnings.simplefilter("ignore")
                prediction = self.regressor.predict(X)
                predictions[i] = prediction.ravel()[0]

            # Update `last_window` values. The first position is discarded and 
            # the new prediction is added at the end.
            last_window = np.append(last_window[1:], prediction)

        return predictions

            
    def predict(
        self,
        steps: int,
        last_window: Optional[pd.Series]=None,
        exog: Optional[Union[pd.Series, pd.DataFrame]]=None
    ) -> pd.Series:
        """
        Predict n steps ahead. It is an recursive process in which, each prediction,
        is used as a predictor for the next step.
        
        Parameters
        ----------
        steps : int
            Number of future steps predicted.
        last_window : pandas Series, default `None`
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in `self.last_window` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s.

        Returns
        -------
        predictions : pandas Series
            Predicted values.
        
        """

        if last_window is None:
            last_window = self.last_window

        check_predict_input(
            forecaster_name  = type(self).__name__,
            steps            = steps,
            fitted           = self.fitted,
            included_exog    = self.included_exog,
            index_type       = self.index_type,
            index_freq       = self.index_freq,
            window_size      = self.window_size_diff,
            last_window      = last_window,
            last_window_exog = None,
            exog             = exog,
            exog_type        = self.exog_type,
            exog_col_names   = self.exog_col_names,
            interval         = None,
            alpha            = None,
            max_steps        = None,
            levels           = None,
            series_col_names = None
        )

        last_window = last_window.iloc[-self.window_size_diff:].copy()

        if exog is not None:
            if isinstance(exog, pd.DataFrame):
                exog = transform_dataframe(
                           df                = exog,
                           transformer       = self.transformer_exog,
                           fit               = False,
                           inverse_transform = False
                       )
            else:
                exog = transform_series(
                           series            = exog,
                           transformer       = self.transformer_exog,
                           fit               = False,
                           inverse_transform = False
                       )
            check_exog_dtypes(exog=exog)
            exog_values = exog.to_numpy()[:steps]
        else:
            exog_values = None
        
        last_window = transform_series(
                          series            = last_window,
                          transformer       = self.transformer_y,
                          fit               = False,
                          inverse_transform = False
                      )
        last_window_values, last_window_index = preprocess_last_window(
                                                    last_window = last_window
                                                )
        if self.differentiation is not None:
            last_window_values = self.differentiator.fit_transform(last_window_values)
            
        predictions = self._recursive_predict(
                          steps       = steps,
                          last_window = last_window_values,
                          exog        = exog_values
                      )
        
        if self.differentiation is not None:
            predictions = self.differentiator.inverse_transform_next_window(predictions)

        predictions = pd.Series(
                          data  = predictions,
                          index = expand_index(
                                      index = last_window_index,
                                      steps = steps
                                  ),
                          name = 'pred'
                      )

        predictions = transform_series(
                          series            = predictions,
                          transformer       = self.transformer_y,
                          fit               = False,
                          inverse_transform = True
                      )
        
        return predictions

    
    def predict_bootstrapping(
        self,
        steps: int,
        last_window: Optional[pd.Series]=None,
        exog: Optional[Union[pd.Series, pd.DataFrame]]=None,
        n_boot: int=250,
        random_state: int=123,
        in_sample_residuals: bool=True,
        binned_residuals: bool=False,
    ) -> pd.DataFrame:
        """
        Generate multiple forecasting predictions using a bootstrapping process. 
        By sampling from a collection of past observed errors (the residuals),
        each iteration of bootstrapping generates a different set of predictions. 
        See the Notes section for more information. 
        
        Parameters
        ----------
        steps : int
            Number of future steps predicted.
        last_window : pandas Series, default `None`
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in `self.last_window` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s.
        n_boot : int, default `500`
            Number of bootstrapping iterations used to estimate predictions.
        random_state : int, default `123`
            Sets a seed to the random generator, so that boot predictions are always 
            deterministic.
        in_sample_residuals : bool, default `True`
            If `True`, residuals from the training data are used as proxy of
            prediction error to create predictions. If `False`, out of sample 
            residuals are used. In the latter case, the user should have
            calculated and stored the residuals within the forecaster (see
            `set_out_sample_residuals()`).
        binned_residuals : bool, default `False`
            If `True`, residuals used in each bootstrapping iteration are selected
            conditioning on the predicted values. If `False`, residuals are selected
            randomly without conditioning on the predicted values.
            **WARNING: This argument is newly introduced and requires special attention.
            It is still experimental and may undergo changes.**
            **New in version 0.12.0**

        Returns
        -------
        boot_predictions : pandas DataFrame
            Predictions generated by bootstrapping.
            Shape: (steps, n_boot)

        Notes
        -----
        More information about prediction intervals in forecasting:
        https://otexts.com/fpp3/prediction-intervals.html#prediction-intervals-from-bootstrapped-residuals
        Forecasting: Principles and Practice (3nd ed) Rob J Hyndman and George Athanasopoulos.

        """

        # TODO: Move to check_predict_input(), validate why it was not there.
        if not in_sample_residuals:
            if not binned_residuals and self.out_sample_residuals is None:
                raise ValueError(
                    ("`forecaster.out_sample_residuals` is `None`. Use "
                     "`in_sample_residuals=True` or method `set_out_sample_residuals()` "
                     "before `predict_interval()`, `predict_bootstrapping()`, "
                     "`predict_quantiles()` or `predict_dist()`.")
                )
            if binned_residuals and self.out_sample_residuals_by_bin is None:
                raise ValueError(
                    ("`forecaster.out_sample_residuals_by_bin` is `None`. Use "
                     "`in_sample_residuals=True` or method `set_out_sample_residuals()` "
                     "before `predict_interval()`, `predict_bootstrapping()`, "
                     "`predict_quantiles()` or `predict_dist()`.")
                )
        
        if last_window is None:
            last_window = self.last_window

        check_predict_input(
            forecaster_name  = type(self).__name__,
            steps            = steps,
            fitted           = self.fitted,
            included_exog    = self.included_exog,
            index_type       = self.index_type,
            index_freq       = self.index_freq,
            window_size      = self.window_size_diff,
            last_window      = last_window,
            last_window_exog = None,
            exog             = exog,
            exog_type        = self.exog_type,
            exog_col_names   = self.exog_col_names,
            interval         = None,
            alpha            = None,
            max_steps        = None,
            levels           = None,
            series_col_names = None
        )

        last_window = last_window.iloc[-self.window_size_diff:]

        if exog is not None:
            if isinstance(exog, pd.DataFrame):
                exog = transform_dataframe(
                           df                = exog,
                           transformer       = self.transformer_exog,
                           fit               = False,
                           inverse_transform = False
                       )
            else:
                exog = transform_series(
                           series            = exog,
                           transformer       = self.transformer_exog,
                           fit               = False,
                           inverse_transform = False
                       )
            exog_values = exog.to_numpy()[:steps]
        else:
            exog_values = None
        
        last_window = transform_series(
                          series            = last_window,
                          transformer       = self.transformer_y,
                          fit               = False,
                          inverse_transform = False
                      )
        last_window_values, last_window_index = preprocess_last_window(
                                                    last_window = last_window
                                                )
        if self.differentiation is not None:
            last_window_values = self.differentiator.fit_transform(last_window_values)

        boot_predictions = np.full(
                               shape      = (steps, n_boot),
                               fill_value = np.nan,
                               dtype      = float
                           )
        rng = np.random.default_rng(seed=random_state)
        seeds = rng.integers(low=0, high=10000, size=n_boot)

        if in_sample_residuals:
            residuals = self.in_sample_residuals
            residuals_by_bin = self.in_sample_residuals_by_bin
        else:
            residuals = self.out_sample_residuals
            residuals_by_bin = self.out_sample_residuals_by_bin

        for i in range(n_boot):
            # In each bootstraping iteration the initial last_window and exog 
            # need to be restored.
            last_window_boot = last_window_values.copy()
            exog_boot = exog_values.copy() if exog is not None else None

            rng = np.random.default_rng(seed=seeds[i])
            if not binned_residuals:
                sampled_residuals = rng.choice(
                                        a       = residuals,
                                        size    = steps,
                                        replace = True
                                    )
            
            for step in range(steps):

                prediction = self._recursive_predict(
                                 steps       = 1,
                                 last_window = last_window_boot,
                                 exog        = exog_boot 
                             )
                if binned_residuals:
                    predicted_bin = (
                        self.binner.transform(prediction.reshape(1, -1)).astype(int)[0][0]
                    )
                    sampled_residual = rng.choice(a=residuals_by_bin[predicted_bin], size=1)
                else:
                    sampled_residual = sampled_residuals[step]

                prediction_with_residual  = prediction + sampled_residual
                boot_predictions[step, i] = prediction_with_residual[0]
                last_window_boot = np.append(
                                       last_window_boot[1:],
                                       prediction_with_residual
                                   )
                if exog is not None:
                    exog_boot = exog_boot[1:]

            if self.differentiation is not None:
                boot_predictions[:, i] = (
                    self.differentiator.inverse_transform_next_window(boot_predictions[:, i])
                )

        boot_predictions = pd.DataFrame(
                               data    = boot_predictions,
                               index   = expand_index(last_window_index, steps=steps),
                               columns = [f"pred_boot_{i}" for i in range(n_boot)]
                           )

        if self.transformer_y:
            for col in boot_predictions.columns:
                boot_predictions[col] = transform_series(
                                            series            = boot_predictions[col],
                                            transformer       = self.transformer_y,
                                            fit               = False,
                                            inverse_transform = True
                                        )

        return boot_predictions


    def predict_interval(
        self,
        steps: int,
        last_window: Optional[pd.Series]=None,
        exog: Optional[Union[pd.Series, pd.DataFrame]]=None,
        interval: list=[5, 95],
        n_boot: int=250,
        random_state: int=123,
        in_sample_residuals: bool=True,
        binned_residuals: bool=False
    ) -> pd.DataFrame:
        """
        Iterative process in which each prediction is used as a predictor
        for the next step, and bootstrapping is used to estimate prediction
        intervals. Both predictions and intervals are returned.
        
        Parameters
        ----------
        steps : int
            Number of future steps predicted.
        last_window : pandas Series, default `None`
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in` self.last_window` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s.
        interval : list, default `[5, 95]`
            Confidence of the prediction interval estimated. Sequence of 
            percentiles to compute, which must be between 0 and 100 inclusive. 
            For example, interval of 95% should be as `interval = [2.5, 97.5]`.
        n_boot : int, default `500`
            Number of bootstrapping iterations used to estimate predictions.
        random_state : int, default `123`
            Sets a seed to the random generator, so that boot predictions are always 
            deterministic.
        in_sample_residuals : bool, default `True`
            If `True`, residuals from the training data are used as proxy of
            prediction error to create predictions. If `False`, out of sample 
            residuals are used. In the latter case, the user should have
            calculated and stored the residuals within the forecaster (see
            `set_out_sample_residuals()`).
        binned_residuals : bool, default `False`
            If `True`, residuals used in each bootstrapping iteration are selected
            conditioning on the predicted values. If `False`, residuals are selected
            randomly without conditioning on the predicted values.
            **WARNING: This argument is newly introduced and requires special attention.
            It is still experimental and may undergo changes.**
            **New in version 0.12.0**

        Returns
        -------
        predictions : pandas DataFrame
            Values predicted by the forecaster and their estimated interval.

            - pred: predictions.
            - lower_bound: lower bound of the interval.
            - upper_bound: upper bound of the interval.

        Notes
        -----
        More information about prediction intervals in forecasting:
        https://otexts.com/fpp2/prediction-intervals.html
        Forecasting: Principles and Practice (2nd ed) Rob J Hyndman and
        George Athanasopoulos.
        
        """
        
        check_interval(interval=interval)

        predictions = self.predict(
                          steps       = steps,
                          last_window = last_window,
                          exog        = exog
                      )

        boot_predictions = self.predict_bootstrapping(
                               steps               = steps,
                               last_window         = last_window,
                               exog                = exog,
                               n_boot              = n_boot,
                               random_state        = random_state,
                               in_sample_residuals = in_sample_residuals,
                               binned_residuals    = binned_residuals
                           )

        interval = np.array(interval)/100
        predictions_interval = boot_predictions.quantile(q=interval, axis=1).transpose()
        predictions_interval.columns = ['lower_bound', 'upper_bound']
        predictions = pd.concat((predictions, predictions_interval), axis=1)

        return predictions


    def predict_quantiles(
        self,
        steps: int,
        last_window: Optional[pd.Series]=None,
        exog: Optional[Union[pd.Series, pd.DataFrame]]=None,
        quantiles: list=[0.05, 0.5, 0.95],
        n_boot: int=250,
        random_state: int=123,
        in_sample_residuals: bool=True,
        binned_residuals: bool=False
    ) -> pd.DataFrame:
        """
        Calculate the specified quantiles for each step. After generating 
        multiple forecasting predictions through a bootstrapping process, each 
        quantile is calculated for each step.
        
        Parameters
        ----------
        steps : int
            Number of future steps predicted.
        last_window : pandas Series, default `None`
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in` self.last_window` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s.
        quantiles : list, default `[0.05, 0.5, 0.95]`
            Sequence of quantiles to compute, which must be between 0 and 1 
            inclusive. For example, quantiles of 0.05, 0.5 and 0.95 should be as 
            `quantiles = [0.05, 0.5, 0.95]`.
        n_boot : int, default `500`
            Number of bootstrapping iterations used to estimate quantiles.
        random_state : int, default `123`
            Sets a seed to the random generator, so that boot quantiles are always 
            deterministic.
        in_sample_residuals : bool, default `True`
            If `True`, residuals from the training data are used as proxy of
            prediction error to create prediction quantiles. If `False`, out of
            sample residuals are used. In the latter case, the user should have
            calculated and stored the residuals within the forecaster (see
            `set_out_sample_residuals()`).
        binned_residuals : bool, default `False`
            If `True`, residuals used in each bootstrapping iteration are selected
            conditioning on the predicted values. If `False`, residuals are selected
            randomly without conditioning on the predicted values.
            **WARNING: This argument is newly introduced and requires special attention.
            It is still experimental and may undergo changes.**
            **New in version 0.12.0**

        Returns
        -------
        predictions : pandas DataFrame
            Quantiles predicted by the forecaster.

        Notes
        -----
        More information about prediction intervals in forecasting:
        https://otexts.com/fpp2/prediction-intervals.html
        Forecasting: Principles and Practice (2nd ed) Rob J Hyndman and
        George Athanasopoulos.
        
        """
        
        check_interval(quantiles=quantiles)

        boot_predictions = self.predict_bootstrapping(
                               steps               = steps,
                               last_window         = last_window,
                               exog                = exog,
                               n_boot              = n_boot,
                               random_state        = random_state,
                               in_sample_residuals = in_sample_residuals,
                               binned_residuals    = binned_residuals
                           )

        predictions = boot_predictions.quantile(q=quantiles, axis=1).transpose()
        predictions.columns = [f'q_{q}' for q in quantiles]

        return predictions


    def predict_dist(
        self,
        steps: int,
        distribution: object,
        last_window: Optional[pd.Series]=None,
        exog: Optional[Union[pd.Series, pd.DataFrame]]=None,
        n_boot: int=250,
        random_state: int=123,
        in_sample_residuals: bool=True,
        binned_residuals: bool=False
    ) -> pd.DataFrame:
        """
        Fit a given probability distribution for each step. After generating 
        multiple forecasting predictions through a bootstrapping process, each 
        step is fitted to the given distribution.
        
        Parameters
        ----------
        steps : int
            Number of future steps predicted.
        distribution : Object
            A distribution object from scipy.stats.
        last_window : pandas Series, default `None`
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).  
            If `last_window = None`, the values stored in` self.last_window` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s.
        n_boot : int, default `500`
            Number of bootstrapping iterations used to estimate predictions.
        random_state : int, default `123`
            Sets a seed to the random generator, so that boot predictions are always 
            deterministic.
        in_sample_residuals : bool, default `True`
            If `True`, residuals from the training data are used as proxy of
            prediction error to create predictions. If `False`, out of sample 
            residuals are used. In the latter case, the user should have
            calculated and stored the residuals within the forecaster (see
            `set_out_sample_residuals()`).
        binned_residuals : bool, default `False`
            If `True`, residuals used in each bootstrapping iteration are selected
            conditioning on the predicted values. If `False`, residuals are selected
            randomly without conditioning on the predicted values.
            **WARNING: This argument is newly introduced and requires special attention.
            It is still experimental and may undergo changes.**
            **New in version 0.12.0**

        Returns
        -------
        predictions : pandas DataFrame
            Distribution parameters estimated for each step.

        """

        boot_samples = self.predict_bootstrapping(
                           steps               = steps,
                           last_window         = last_window,
                           exog                = exog,
                           n_boot              = n_boot,
                           random_state        = random_state,
                           in_sample_residuals = in_sample_residuals,
                           binned_residuals    = binned_residuals
                       )       

        param_names = [p for p in inspect.signature(distribution._pdf).parameters
                       if not p=='x'] + ["loc","scale"]
        param_values = np.apply_along_axis(
                           lambda x: distribution.fit(x),
                           axis = 1,
                           arr  = boot_samples
                       )
        predictions = pd.DataFrame(
                          data    = param_values,
                          columns = param_names,
                          index   = boot_samples.index
                      )

        return predictions

    
    def set_params(
        self, 
        params: dict
    ) -> None:
        """
        Set new values to the parameters of the scikit learn model stored in the
        forecaster.
        
        Parameters
        ----------
        params : dict
            Parameters values.

        Returns
        -------
        None
        
        """

        self.regressor = clone(self.regressor)
        self.regressor.set_params(**params)


    def set_fit_kwargs(
        self, 
        fit_kwargs: dict
    ) -> None:
        """
        Set new values for the additional keyword arguments passed to the `fit` 
        method of the regressor.
        
        Parameters
        ----------
        fit_kwargs : dict
            Dict of the form {"argument": new_value}.

        Returns
        -------
        None
        
        """

        self.fit_kwargs = check_select_fit_kwargs(self.regressor, fit_kwargs=fit_kwargs)

        
    def set_lags(
        self, 
        lags: Union[int, list, np.ndarray, range]
    ) -> None:
        """
        Set new value to the attribute `lags`. Attributes `max_lag`, 
        `window_size` and  `window_size_diff` are also updated.
        
        Parameters
        ----------
        lags : int, list, numpy ndarray, range
            Lags used as predictors. Index starts at 1, so lag 1 is equal to t-1.

            - `int`: include lags from 1 to `lags` (included).
            - `list`, `1d numpy ndarray` or `range`: include only lags present in 
            `lags`, all elements must be int.

        Returns
        -------
        None
        
        """
        
        self.lags = initialize_lags(type(self).__name__, lags)
        self.max_lag = max(self.lags)
        self.window_size = max(self.lags)
        self.window_size_diff = max(self.lags)
        if self.differentiation is not None:
            self.window_size_diff += self.differentiation        
        

    def set_out_sample_residuals(
        self, 
        residuals: Union[pd.Series, np.ndarray],
        y_pred: Optional[Union[pd.Series, np.ndarray]]=None,
        append: bool=True,
        transform: bool=True,
        random_state: int=123
    )-> None:
        """
        Set new values to the attribute `out_sample_residuals`. Out of sample
        residuals are meant to be calculated using observations that did not
        participate in the training process. If `y_pred` is provided, residuals
        are binned according to the predicted value they are associated with. If
        `y_pred` is `None`, residuals are stored without binning. Only up to 200
        residuals are stored per bin.
        
        Parameters
        ----------
        residuals : pandas Series, numpy ndarray
            Values of residuals. If `y_pred` is `None`, at most 1000 values are
            stored. If `y_pred` is not `None`, at most 200 * n_bins values are
            stored, where `n_bins` is the number of bins used in `self.binner`.
        y_pred : pandas Series, numpy ndarray, default `None`
            Predicted values of the time series from which the residuals have been
            calculated. This argument is used to bin residuals according to the
            predicted values. `y_pred` and `residuals` must be of the same class
            (both pandas Series or both numpy ndarray) must have the same length
            and, if they are pandas Series, the same index. 
            
            - If `y_pred` is `None`, residuals are not binned.
            - If affter binning, a bin has more than 200 residuals, only a random
                sample of 200 residuals is stored.
            - If affter binning, a bin binning is empty, it is filled with a
            random sample of residuals from other bins. This is done to ensure
            that all bins have at least one residual and can be used in the
            prediction process.
            **New in version 0.12.0**
        append : bool, default `True`
            If `True`, new residuals are added to the once already stored in the
            forecaster. Once the limit of 200 values per bin is reached, no more values
            are appended. If False, stored residuals are overwritten with the new
            residuals.
        transform : bool, default `True`
            If `True`, new residuals are transformed using self.transformer_y.
        random_state : int, default `123`
            Sets a seed to the random sampling for reproducible output.

        Returns
        -------
        None

        """

        if not isinstance(residuals, (np.ndarray, pd.Series)):
            raise TypeError(
                (f"`residuals` argument must be `numpy ndarray` or `pandas Series`, "
                 f"but found {type(residuals)}.")
            )

        if not isinstance(y_pred, (np.ndarray, pd.Series, type(None))):
            raise TypeError(
                (f"`y_pred` argument must be `numpy ndarray`, `pandas Series` or `None`, "
                 f"but found {type(y_pred)}.")
            )

        if y_pred is not None and len(residuals) != len(y_pred):
            raise ValueError(
                (f"`residuals` and `y_pred` must have the same length, but found "
                 f"{len(residuals)} and {len(y_pred)}.")
            )

        if isinstance(residuals, pd.Series) and isinstance(y_pred, pd.Series):
            if not residuals.index.equals(y_pred.index):
                raise ValueError(
                    (f"`residuals` and `y_pred` must have the same index, but found "
                     f"{residuals.index} and {y_pred.index}.")
                )

        if y_pred is not None and not self.fitted:
            raise NotFittedError(
                ("This forecaster is not fitted yet. Call `fit` with appropriate "
                 "arguments before using `set_out_sample_residuals()`.")
            )

        if isinstance(residuals, np.ndarray):
            residuals = pd.Series(residuals, name='residuals')
        else:
            residuals = residuals.rename('residuals').reset_index(drop=True)

        if isinstance(y_pred, np.ndarray):
            y_pred = pd.Series(y_pred, name='prediction')
        elif isinstance(y_pred, pd.Series):
            y_pred = y_pred.rename('prediction').reset_index(drop=True)

        if not transform and self.transformer_y is not None:
            warnings.warn(
                (f"Argument `transform` is set to `False` but forecaster was trained "
                 f"using a transformer {self.transformer_y}. Ensure that the new residuals "
                 f"are already transformed or set `transform=True`.")
            )

        if transform and self.transformer_y is not None:
            warnings.warn(
                (f"Residuals will be transformed using the same transformer used "
                 f"when training the forecaster ({self.transformer_y}). Ensure that the "
                 f"new residuals are on the same scale as the original time series.")
            )

            residuals = transform_series(
                            series            = residuals,
                            transformer       = self.transformer_y,
                            fit               = False,
                            inverse_transform = False
                        ).to_numpy()
        
        if y_pred is None:
            # Residuals are not binned.
            if len(residuals) > 1000:
                rng = np.random.default_rng(seed=random_state)
                residuals = rng.choice(a=residuals, size=1000, replace=False)
            if append and self.out_sample_residuals is not None:
                free_space = max(0, 1000 - len(self.out_sample_residuals))
                if len(residuals) < free_space:
                    residuals = np.hstack((
                                    self.out_sample_residuals,
                                    residuals
                                ))
                else:
                    residuals = np.hstack((
                                    self.out_sample_residuals,
                                    residuals[:free_space]
                                ))
            self.out_sample_residuals = residuals
        else:
            # Residuals are binned according to the predicted values.
            data = pd.merge(
                       residuals,
                       y_pred,
                       left_index  = True,
                       right_index = True
                   )
            data['bin'] = self.binner.transform(data[['prediction']].to_numpy()).astype(int)
            residuals_by_bin = data.groupby('bin')['residuals'].apply(np.array).to_dict()

            if append and self.out_sample_residuals_by_bin is not None:
                for k, v in residuals_by_bin.items():
                    if k in self.out_sample_residuals_by_bin:
                        free_space = max(0, 200 - len(self.out_sample_residuals_by_bin[k]))
                        if len(v) < free_space:
                            self.out_sample_residuals_by_bin[k] = np.hstack((
                                self.out_sample_residuals_by_bin[k],
                                v
                            ))
                        else:
                            self.out_sample_residuals_by_bin[k] = np.hstack((
                                self.out_sample_residuals_by_bin[k],
                                v[:free_space]
                            ))
                    else:
                        self.out_sample_residuals_by_bin[k] = v
            else:
                self.out_sample_residuals_by_bin = residuals_by_bin

            for k, v in self.out_sample_residuals_by_bin.items():
                rng = np.random.default_rng(seed=123)
                if len(v) > 200:
                    # Only up to 200 residuals are stored per bin
                    sample = rng.choice(a=v, size=200, replace=False)
                    self.out_sample_residuals_by_bin[k] = sample
                
            self.out_sample_residuals = np.concatenate(list(
                                            self.out_sample_residuals_by_bin.values()
                                        ))

            for k in self.in_sample_residuals_by_bin.keys():
                if k not in self.out_sample_residuals_by_bin:
                    self.out_sample_residuals_by_bin[k] = np.array([])
            
            empty_bins = [k for k, v in self.out_sample_residuals_by_bin.items() if len(v) == 0]
            if empty_bins:
                warnings.warn(
                    (f"The following bins have no out of sample residuals: {empty_bins}. "
                     f"No predicted values fall in the interval "
                     f"{[self.binner_intervals[bin] for bin in empty_bins]}. "
                     f"Empty bins will be filled with a random sample of residuals from "
                     f"the other bins.")
                )
                for k in empty_bins:
                    rng = np.random.default_rng(seed=123)
                    self.out_sample_residuals_by_bin[k] = rng.choice(
                                                              a       = self.out_sample_residuals,
                                                              size    = 200,
                                                              replace = True
                                                          )


    def get_feature_importances(
        self,
        sort_importance: bool=True
    ) -> pd.DataFrame:
        """
        Return feature importances of the regressor stored in the forecaster.
        Only valid when regressor stores internally the feature importances in the
        attribute `feature_importances_` or `coef_`. Otherwise, returns `None`.

        Parameters
        ----------
        sort_importance: bool, default `True`
            If `True`, sorts the feature importances in descending order.

        Returns
        -------
        feature_importances : pandas DataFrame
            Feature importances associated with each predictor.

        """

        if not self.fitted:
            raise NotFittedError(
                ("This forecaster is not fitted yet. Call `fit` with appropriate "
                 "arguments before using `get_feature_importances()`.")
            )

        if isinstance(self.regressor, Pipeline):
            estimator = self.regressor[-1]
        else:
            estimator = self.regressor

        if hasattr(estimator, 'feature_importances_'):
            feature_importances = estimator.feature_importances_
        elif hasattr(estimator, 'coef_'):
            feature_importances = estimator.coef_
        else:
            warnings.warn(
                (f"Impossible to access feature importances for regressor of type "
                 f"{type(estimator)}. This method is only valid when the "
                 f"regressor stores internally the feature importances in the "
                 f"attribute `feature_importances_` or `coef_`.")
            )
            feature_importances = None

        if feature_importances is not None:
            feature_importances = pd.DataFrame({
                                      'feature': self.X_train_col_names,
                                      'importance': feature_importances
                                  })
            if sort_importance:
                feature_importances = feature_importances.sort_values(
                                          by='importance', ascending=False
                                      )

        return feature_importances


# Unit test predict_bootstrapping ForecasterAutoreg
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler

from skforecast.ForecasterAutoreg import ForecasterAutoreg
from skforecast.preprocessing import TimeSeriesDifferentiator

# Fixtures
from .fixtures_ForecasterAutoreg import y
from .fixtures_ForecasterAutoreg import exog
from .fixtures_ForecasterAutoreg import exog_predict
from .fixtures_ForecasterAutoreg import data # to test results when using differentiation


def test_predict_bootstrapping_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)

    err_msg = re.escape(
        ("This Forecaster instance is not fitted yet. Call `fit` with "
         "appropriate arguments before using predict.")
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1)


def test_predict_bootstrapping_ValueError_when_out_sample_residuals_is_None():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals is None.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)))

    err_msg = re.escape(
        ("`forecaster.out_sample_residuals` is `None`. Use "
         "`in_sample_residuals=True` or method `set_out_sample_residuals()` "
         "before `predict_interval()`, `predict_bootstrapping()`, "
         "`predict_quantiles()` or `predict_dist()`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, in_sample_residuals=False)


def test_predict_bootstrapping_ValueError_when_out_sample_residuals_by_bin_is_None():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals is None.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)))

    err_msg = re.escape(
        ("`forecaster.out_sample_residuals_by_bin` is `None`. Use "
         "`in_sample_residuals=True` or method `set_out_sample_residuals()` "
         "before `predict_interval()`, `predict_bootstrapping()`, "
         "`predict_quantiles()` or `predict_dist()`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, in_sample_residuals=False,
                                         binned_residuals=True)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_1_in_sample_residuals_is_True():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    1 step ahead is predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=y, exog=exog)
    expected = pd.DataFrame(
                   data    = np.array([[0.83836963, 0.54888446, 1.16537933, 0.69662898]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=51)
               )
    results = forecaster.predict_bootstrapping(steps=1, n_boot=4, exog=exog_predict, in_sample_residuals=True)

    pd.testing.assert_frame_equal(expected, results)
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_True():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    2 steps ahead are predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=y, exog=exog)
    expected = pd.DataFrame(
                   data    = np.array(
                                [[ 0.83836963,  0.54888446,  1.16537933,  0.69662898],
                                [ 0.21874597, -0.0376708 ,  0.29184038,  0.5833093 ]]
                            ),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    results = forecaster.predict_bootstrapping(steps=2, n_boot=4, exog=exog_predict, in_sample_residuals=True)

    pd.testing.assert_frame_equal(expected, results)
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_1_in_sample_residuals_is_False():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    1 step ahead is predicted with exog using out-sample residuals.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=y, exog=exog)
    forecaster.out_sample_residuals = forecaster.in_sample_residuals
    expected = pd.DataFrame(
                   data    = np.array([[0.83836963, 0.54888446, 1.16537933, 0.69662898]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=51)
               )
    results = forecaster.predict_bootstrapping(steps=1, n_boot=4, exog=exog_predict, in_sample_residuals=False)

    pd.testing.assert_frame_equal(expected, results)
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_False():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    2 steps ahead are predicted with exog using out-sample residuals.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=y, exog=exog)
    forecaster.out_sample_residuals = forecaster.in_sample_residuals
    expected = pd.DataFrame(
                   data    = np.array([[ 0.83836963,  0.54888446,  1.16537933,  0.69662898],
                                       [ 0.21874597, -0.0376708 ,  0.29184038,  0.5833093 ]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    results = forecaster.predict_bootstrapping(steps=2, n_boot=4, exog=exog_predict, in_sample_residuals=False)

    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer_fixed_to_zero():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included, both
    inputs are transformed and in-sample residuals are fixed to 0.
    """
    forecaster = ForecasterAutoreg(
                     regressor        = LinearRegression(),
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )
    forecaster.fit(y=y, exog=exog)
    forecaster.in_sample_residuals = np.full_like(forecaster.in_sample_residuals, fill_value=0)
    results = forecaster.predict_bootstrapping(steps=2, exog=exog_predict, n_boot=4, in_sample_residuals=True)
    expected = pd.DataFrame(
                   data = np.array([[0.67998879, 0.67998879, 0.67998879, 0.67998879],
                                   [0.46135022, 0.46135022, 0.46135022, 0.46135022]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """

    forecaster = ForecasterAutoreg(
                     regressor        = LinearRegression(),
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster.predict_bootstrapping(steps=2, exog=exog_predict, n_boot=4, in_sample_residuals=True)
    expected = pd.DataFrame(
                   data    = np.array(
                                 [[ 0.83836963,  0.54888446,  1.16537933,  0.69662898],
                                  [ 0.21874597, -0.0376708 ,  0.29184038,  0.5833093 ]]
                             ),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
                )

    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_and_differentiation_is_1():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    differentiation is 1.
    """
    # Data differentiated
    diferenciator = TimeSeriesDifferentiator(order=1)
    data_diff = diferenciator.fit_transform(data.to_numpy())
    data_diff = pd.Series(data_diff, index=data.index).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'
    steps = len(data.loc[end_train:])

    forecaster_1 = ForecasterAutoreg(regressor=LinearRegression(), lags=15)
    forecaster_1.fit(y=data_diff.loc[:end_train], exog=exog_diff.loc[:end_train])
    boot_predictions_diff = forecaster_1.predict_bootstrapping(
                                steps=steps,
                                exog=exog_diff.loc[end_train:],
                                n_boot=10
                            )
    last_value_train = data.loc[:end_train].iloc[[-1]]
    boot_predictions_1 = boot_predictions_diff.copy()
    boot_predictions_1.loc[last_value_train.index[0]] = last_value_train.values[0]
    boot_predictions_1 = boot_predictions_1.sort_index()
    boot_predictions_1 = boot_predictions_1.cumsum(axis=0).iloc[1:,]
    boot_predictions_1 = boot_predictions_1.asfreq('MS')
    
    forecaster_2 = ForecasterAutoreg(regressor=LinearRegression(), lags=15, differentiation=1)
    forecaster_2.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    boot_predictions_2 = forecaster_2.predict_bootstrapping(
                            steps  = steps,
                            exog   = exog_diff.loc[end_train:],
                            n_boot = 10
                        )
    
    pd.testing.assert_frame_equal(boot_predictions_1, boot_predictions_2)


# Unit test save_forecaster and load_forecaster
# ==============================================================================
import os
import re
import joblib
import pytest
import inspect
import warnings
import numpy as np
import pandas as pd
import skforecast
from skforecast.ForecasterAutoreg import ForecasterAutoreg
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom
from skforecast.utils import save_forecaster
from skforecast.utils import load_forecaster
from skforecast.exceptions import SkforecastVersionWarning
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler

def create_predictors(y): # pragma: no cover
    """
    Create first 2 lags of a time series.
    """
    
    lags = y[-1:-3:-1]
    
    return lags

def custom_weights(y): # pragma: no cover
    """
    """
    return np.ones(len(y))

def custom_weights2(y): # pragma: no cover
    """
    """
    return np.arange(1, len(y)+1)


def test_save_and_load_forecaster_persistence():
    """ 
    Test if a loaded forecaster is exactly the same as the original one.
    """
    forecaster = ForecasterAutoreg(regressor=LinearRegression(), lags=3, transformer_y=StandardScaler())
    rng = np.random.default_rng(12345)
    y = pd.Series(rng.normal(size=100))
    forecaster.fit(y=y)
    save_forecaster(forecaster=forecaster, file_name='forecaster.joblib', verbose=True)
    forecaster_loaded = load_forecaster(file_name='forecaster.joblib', verbose=True)
    os.remove('forecaster.joblib')

    for key in vars(forecaster).keys():
        attribute_forecaster = forecaster.__getattribute__(key)
        attribute_forecaster_loaded = forecaster_loaded.__getattribute__(key)

        if key in ['regressor', 'binner', 'transformer_y', 'transformer_exog']:
            assert joblib.hash(attribute_forecaster) == joblib.hash(attribute_forecaster_loaded)
        elif isinstance(attribute_forecaster, (np.ndarray, pd.Series, pd.DataFrame, pd.Index)):
            assert (attribute_forecaster == attribute_forecaster_loaded).all()
        elif isinstance(attribute_forecaster, dict):
            assert attribute_forecaster.keys() == attribute_forecaster_loaded.keys()
            for k in attribute_forecaster.keys():
                if isinstance(attribute_forecaster[k], (np.ndarray, pd.Series, pd.DataFrame, pd.Index)):
                    assert (attribute_forecaster[k] == attribute_forecaster_loaded[k]).all()
                else:
                    assert attribute_forecaster[k] == attribute_forecaster_loaded[k]
        else:
            assert attribute_forecaster == attribute_forecaster_loaded


def test_save_and_load_forecaster_SkforecastVersionWarning():
    """ 
    Test warning used to notify that the skforecast version installed in the 
    environment differs from the version used to create the forecaster.
    """
    forecaster = ForecasterAutoreg(regressor=LinearRegression(), lags=3)
    rng = np.random.default_rng(123)
    y = pd.Series(rng.normal(size=100))
    forecaster.fit(y=y)
    forecaster.skforecast_version = '0.0.0'
    save_forecaster(forecaster=forecaster, file_name='forecaster.joblib', verbose=False)

    warn_msg = re.escape(
        (f"The skforecast version installed in the environment differs "
         f"from the version used to create the forecaster.\n"
         f"    Installed Version  : {skforecast.__version__}\n"
         f"    Forecaster Version : 0.0.0\n"
         f"This may create incompatibilities when using the library.")
    )
    with pytest.warns(SkforecastVersionWarning, match = warn_msg):
        load_forecaster(file_name='forecaster.joblib', verbose=False)
        os.remove('forecaster.joblib')


@pytest.mark.parametrize("weight_func", 
                         [custom_weights, 
                          {'serie_1': custom_weights, 
                           'serie_2': custom_weights2}], 
                         ids = lambda func : f'type: {type(func)}')
def test_save_forecaster_save_custom_functions(weight_func):
    """ 
    Test if custom functions are saved correctly.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors,
                     window_size    = 3,
                     weight_func    = weight_func
                 )
    save_forecaster(forecaster=forecaster, file_name='forecaster.joblib', 
                    save_custom_functions=True)
    os.remove('forecaster.joblib')

    fun_predictors_file = forecaster.fun_predictors.__name__ + '.py'
    assert os.path.exists(fun_predictors_file)
    with open(fun_predictors_file, 'r') as file:
        assert inspect.getsource(forecaster.fun_predictors) == file.read()
    os.remove(fun_predictors_file)
    
    weight_functions = weight_func.values() if isinstance(weight_func, dict) else [weight_func]
    for weight_func in weight_functions:
        weight_func_file = weight_func.__name__ + '.py'
        assert os.path.exists(weight_func_file)
        with open(weight_func_file, 'r') as file:
            assert inspect.getsource(weight_func) == file.read()
        os.remove(weight_func_file)


@pytest.mark.parametrize("weight_func", 
                         [None,
                          custom_weights, 
                          {'serie_1': custom_weights, 
                           'serie_2': custom_weights2}], 
                         ids = lambda func : f'func: {func}')
def test_save_forecaster_warning_dont_save_custom_functions(weight_func):
    """ 
    Test UserWarning when custom functions are not saved.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors,
                     window_size    = 3,
                     weight_func    = weight_func
                 )

    warn_msg = re.escape(
        ("Custom functions used to create predictors or weights are not saved. "
         "To save them, set `save_custom_functions` to `True`.")
    )
    with pytest.warns(UserWarning, match = warn_msg):
        save_forecaster(forecaster=forecaster, file_name='forecaster.joblib', 
                        save_custom_functions=False)
        os.remove('forecaster.joblib')


# Unit test load_demo_dataset
# ==============================================================================
import pandas as pd
from skforecast.datasets import load_demo_dataset


def test_load_demo_dataset():
    """
    Test load_demo_dataset function.
    """
    df = load_demo_dataset()
    
    assert isinstance(df, pd.Series)
    assert df.index.freq == 'MS'
    assert df.index.is_monotonic_increasing
    assert df.index[0] == pd.Timestamp('1991-07-01')
    assert df.index[-1] == pd.Timestamp('2008-06-01')
    assert df.shape == (204,)


from .model_selection_multiseries import (
    backtesting_forecaster_multiseries,
    grid_search_forecaster_multiseries,
    random_search_forecaster_multiseries,
    bayesian_search_forecaster_multiseries,
    backtesting_forecaster_multivariate,
    grid_search_forecaster_multivariate,
    random_search_forecaster_multivariate,
    bayesian_search_forecaster_multivariate,
    select_features_multiseries
)


# Unit test fit ForecasterAutoregMultiSeriesCustom
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler


def create_predictors(y): # pragma: no cover
    """
    Create first 3 lags of a time series.
    """
    lags = y[-1:-4:-1]

    return lags


def test_fit_correct_dict_create_series_weights_weight_func_transformer_series():
    """
    Test fit method creates correctly all the auxiliary dicts, series_weights_,
    weight_func_, transformer_series_.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10)), 
                           'l3': pd.Series(np.arange(10))})
                    
    series.index = pd.DatetimeIndex(
                       ['2022-01-04', '2022-01-05', '2022-01-06', 
                        '2022-01-07', '2022-01-08', '2022-01-09', 
                        '2022-01-10', '2022-01-11', '2022-01-12', 
                        '2022-01-13'], dtype='datetime64[ns]', freq='D' 
                   )

    def custom_weights(index):
        """
        Return 0 if index is between '2022-01-08' and '2022-01-10', 1 otherwise.
        """
        weights = np.where(
                      (index >= '2022-01-08') & (index <= '2022-01-10'),
                      0,
                      1
                  )
        
        return weights

    transformer_series = {'l1': StandardScaler()}
    weight_func = {'l2': custom_weights}    
    series_weights = {'l1': 3., 'l3': 0.5, 'l4': 2.}

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = transformer_series,
                     weight_func        = weight_func,
                     series_weights     = series_weights
                 )
    
    forecaster.fit(series=series, store_in_sample_residuals=False)

    expected_transformer_series_ = {'l1': forecaster.transformer_series_['l1'], 'l2': None, 'l3': None}
    expected_weight_func_ = {'l1': lambda index: np.ones_like(index, dtype=float), 'l2': custom_weights, 'l3': lambda index: np.ones_like(index, dtype=float)}
    expected_series_weights_ = {'l1': 3., 'l2': 1., 'l3': 0.5}

    assert forecaster.transformer_series_ == expected_transformer_series_
    assert forecaster.weight_func_.keys() == expected_weight_func_.keys()
    for key in forecaster.weight_func_.keys():
        assert forecaster.weight_func_[key].__code__.co_code == expected_weight_func_[key].__code__.co_code
    assert forecaster.series_weights_ == expected_series_weights_

    forecaster.fit(series=series[['l1', 'l2']], store_in_sample_residuals=False)

    expected_transformer_series_ = {'l1': forecaster.transformer_series_['l1'], 'l2': None}
    expected_weight_func_ = {'l1': lambda index: np.ones_like(index, dtype=float), 'l2': custom_weights}
    expected_series_weights_ = {'l1': 3., 'l2': 1.}

    assert forecaster.transformer_series_ == expected_transformer_series_
    assert forecaster.weight_func_.keys() == expected_weight_func_.keys()
    for key in forecaster.weight_func_.keys():
        assert forecaster.weight_func_[key].__code__.co_code == expected_weight_func_[key].__code__.co_code
    assert forecaster.series_weights_ == expected_series_weights_


def test_forecaster_DatetimeIndex_index_freq_stored():
    """
    Test serie_with_DatetimeIndex.index.freqstr is stored in forecaster.index_freq.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(5)), 
                           '2': pd.Series(np.arange(5))})

    series.index = pd.date_range(start='2022-01-01', periods=5, freq='1D')

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series)
    expected = series.index.freqstr
    results = forecaster.index_freq

    assert results == expected


def test_forecaster_index_step_stored():
    """
    Test serie without DatetimeIndex, step is stored in forecaster.index_freq.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(5)), 
                           '2': pd.Series(np.arange(5))})
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series)
    expected = series.index.step
    results = forecaster.index_freq

    assert results == expected


def test_fit_in_sample_residuals_stored():
    """
    Test that values of in_sample_residuals are stored after fitting
    when `store_in_sample_residuals=True`.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(5)), 
                           '2': pd.Series(np.arange(5))})

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    expected = {'1': np.array([-4.4408921e-16, 0.0000000e+00]),
                '2': np.array([0., 0.])}
    results = forecaster.in_sample_residuals

    assert isinstance(results, dict)
    assert np.all(isinstance(x, np.ndarray) for x in results.values())
    assert results.keys() == expected.keys()
    assert np.all(np.all(np.isclose(results[k], expected[k])) for k in expected.keys())


def test_fit_same_residuals_when_residuals_greater_than_1000():
    """
    Test fit return same residuals when residuals len is greater than 1000.
    Testing with two different forecaster. Residuals shouldn't be more than 
    1000 values.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(1010)), 
                           '2': pd.Series(np.arange(1010))})

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results_1 = forecaster.in_sample_residuals
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results_2 = forecaster.in_sample_residuals

    assert isinstance(results_1, dict)
    assert np.all(isinstance(x, np.ndarray) for x in results_1.values())
    assert isinstance(results_2, dict)
    assert np.all(isinstance(x, np.ndarray) for x in results_2.values())
    assert results_1.keys() == results_2.keys()
    assert np.all(len(results_1[k] == 1000) for k in results_1.keys())
    assert np.all(len(results_2[k] == 1000) for k in results_2.keys())
    assert np.all(np.all(results_1[k] == results_2[k]) for k in results_2.keys())


def test_fit_in_sample_residuals_not_stored():
    """
    Test that values of in_sample_residuals are not stored after fitting
    when `store_in_sample_residuals=False`.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(5)), 
                           '2': pd.Series(np.arange(5))})

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series, store_in_sample_residuals=False)
    expected = {'1': None, '2': None}
    results = forecaster.in_sample_residuals

    assert isinstance(results, dict)
    assert results.keys() == expected.keys()
    assert np.all(results[k] == expected[k] for k in expected.keys())


def test_fit_last_window_stored():
    """
    Test that values of last window are stored after fitting.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(5), dtype=float), 
                           '2': pd.Series(np.arange(5), dtype=float)})

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series)

    expected = {
        '1': pd.Series(
                 data  = np.array([2., 3., 4.]),
                 index = pd.RangeIndex(start=2, stop=5, step=1),
                 name  = '1',
                 dtype = float
             ),
        '2': pd.Series(
                 data  = np.array([2., 3., 4.]),
                 index = pd.RangeIndex(start=2, stop=5, step=1),
                 name  = '2',
                 dtype = float
             )
    }

    for k in forecaster.last_window.keys():
        pd.testing.assert_series_equal(forecaster.last_window[k], expected[k])


@pytest.mark.parametrize("encoding, encoding_mapping", 
                         [('ordinal'         , {'1': 0, '2': 1}), 
                          ('ordinal_category', {'1': 0, '2': 1}),
                          ('onehot'          , {'1': 0, '2': 1})], 
                         ids = lambda dt : f'encoding, mapping: {dt}')
def test_fit_encoding_mapping(encoding, encoding_mapping):
    """
    Test the encoding mapping of _create_train_X_y.
    """
    series = pd.DataFrame({'1': pd.Series(np.arange(7, dtype=float)), 
                           '2': pd.Series(np.arange(7, dtype=float))})

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors,
                     window_size    = 3,
                     encoding       = encoding
                 )
    forecaster.fit(series=series, suppress_warnings=True)
    
    assert forecaster.encoding_mapping == encoding_mapping


# Unit test predict_quantiles ForecasterAutoregMultiSeries
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregMultiSeries import ForecasterAutoregMultiSeries
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression

# Fixtures
from .fixtures_ForecasterAutoregMultiSeries import series
from .fixtures_ForecasterAutoregMultiSeries import exog
from .fixtures_ForecasterAutoregMultiSeries import exog_predict

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


@pytest.mark.parametrize("level", 
                         ['1', ['1']], 
                         ids=lambda lvl: f'level: {lvl}')
def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer(level):
    """
    Test output of predict_quantiles when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed. Single level.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )
    
    forecaster.fit(series=series, exog=exog)
    results = forecaster.predict_quantiles(
                  steps               = 2,
                  quantiles           = [0.05, 0.55, 0.95],
                  levels              = level,
                  exog                = exog_predict,
                  n_boot              = 4,
                  in_sample_residuals = True,
                  suppress_warnings   = True
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.20133572, 0.40460928, 0.53761911],
                                       [0.07739771, 0.19454095, 0.45374409]]),
                   columns = ['1_q_0.05', '1_q_0.55', '1_q_0.95'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("levels", 
                         [['1', '2'], None], 
                         ids=lambda lvl: f'levels: {lvl}')
def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer(levels):
    """
    Test output of predict_quantiles when regressor is LinearRegression,
    2 steps are predicted, using out-sample residuals, exog is included and both
    inputs are transformed. Multiple levels.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )
    
    forecaster.fit(series=series, exog=exog)
    forecaster.out_sample_residuals = forecaster.in_sample_residuals
    results = forecaster.predict_quantiles(
                  steps               = 2,
                  quantiles           = [0.05, 0.55, 0.95],
                  levels              = levels,
                  exog                = exog_predict,
                  n_boot              = 4,
                  in_sample_residuals = False
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.20133572, 0.40460928, 0.53761911, 0.27660015, 0.61593304, 0.7616801 ],
                                       [0.07739771, 0.19454095, 0.45374409, 0.18601931, 0.40403914, 0.70461714]]),
                   columns = ['1_q_0.05', '1_q_0.55', '1_q_0.95', '2_q_0.05', '2_q_0.55', '2_q_0.95'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )

    pd.testing.assert_frame_equal(results, expected)


# Unit test backtesting ForecasterEquivalentDate
# ==============================================================================
import numpy as np
import pandas as pd
from skforecast.ForecasterBaseline import ForecasterEquivalentDate
from skforecast.model_selection import backtesting_forecaster

# Fixtures
from .fixtures_ForecasterEquivalentDate import y


def test_backtesting_with_ForecasterEquivalentDate():
    """
    Test backtesting with ForecasterEquivalentDate.
    """
    forecaster = ForecasterEquivalentDate(
                     offset    = pd.DateOffset(days=10),
                     n_offsets = 2 
                 )

    metric, predictions = backtesting_forecaster(
        forecaster         = forecaster,
        y                  = y,
        initial_train_size = 30,
        steps              = 5,
        metric             = 'mean_absolute_error',
        refit              = True,
        verbose            = False,
        n_jobs             = 'auto'
    )

    expected = pd.DataFrame(
        data    = np.array([0.48878949, 0.78924075, 0.58151378, 0.3353507 , 0.56024382,
                            0.53047716, 0.27214019, 0.20185749, 0.41263271, 0.58140185,
                            0.36325295, 0.64156648, 0.57765904, 0.5523543 , 0.57413684,
                            0.31761006, 0.39406999, 0.56082619, 0.61893703, 0.5664064 ]),
        index   = pd.date_range(start='2000-01-31', periods=20, freq='D'),
        columns = ['pred']
    )

    assert metric == 0.2537094475
    pd.testing.assert_frame_equal(predictions, expected)